import datetime
from math import acos, asin, atan2, cos, degrees, fabs, radians, sin, sqrt, tan
from typing import Dict, Optional, Tuple, Union

try:
    import zoneinfo
except ImportError:
    from backports import zoneinfo

from astral import (
    Depression,
    Minutes,
    Observer,
    SunDirection,
    TimePeriod,
    now,
    refraction_at_zenith,
    today,
)
from astral.julian import (
    julianday,
    julianday_to_juliancentury,
)

__all__ = [
    "sun",
    "dawn",
    "sunrise",
    "noon",
    "midnight",
    "sunset",
    "dusk",
    "daylight",
    "night",
    "twilight",
    "blue_hour",
    "golden_hour",
    "rahukaalam",
    "zenith",
    "azimuth",
    "elevation",
    "time_at_elevation",
]


# Using 32 arc minutes as sun's apparent diameter
SUN_APPARENT_RADIUS = 32.0 / (60.0 * 2.0)


# region Backend
def minutes_to_timedelta(minutes: float) -> datetime.timedelta:
    """Convert a floating point number of minutes to a
    :class:`~datetime.timedelta`
    """
    d = int(minutes / 1440)
    minutes = minutes - (d * 1440)
    minutes = minutes * 60
    s = int(minutes)
    sfrac = minutes - s
    us = int(sfrac * 1_000_000)

    return datetime.timedelta(days=d, seconds=s, microseconds=us)


def geom_mean_long_sun(juliancentury: float) -> float:
    """Calculate the geometric mean longitude of the sun"""
    l0 = 280.46646 + juliancentury * (36000.76983 + 0.0003032 * juliancentury)
    return l0 % 360.0


def geom_mean_anomaly_sun(juliancentury: float) -> float:
    """Calculate the geometric mean anomaly of the sun"""
    return 357.52911 + juliancentury * (35999.05029 - 0.0001537 * juliancentury)


def eccentric_location_earth_orbit(juliancentury: float) -> float:
    """Calculate the eccentricity of Earth's orbit"""
    return 0.016708634 - juliancentury * (0.000042037 + 0.0000001267 * juliancentury)


def sun_eq_of_center(juliancentury: float) -> float:
    """Calculate the equation of the center of the sun"""
    m = geom_mean_anomaly_sun(juliancentury)

    mrad = radians(m)
    sinm = sin(mrad)
    sin2m = sin(mrad + mrad)
    sin3m = sin(mrad + mrad + mrad)

    c = (
        sinm * (1.914602 - juliancentury * (0.004817 + 0.000014 * juliancentury))
        + sin2m * (0.019993 - 0.000101 * juliancentury)
        + sin3m * 0.000289
    )

    return c


def sun_true_long(juliancentury: float) -> float:
    """Calculate the sun's true longitude"""
    l0 = geom_mean_long_sun(juliancentury)
    c = sun_eq_of_center(juliancentury)

    return l0 + c


def sun_true_anomoly(juliancentury: float) -> float:
    """Calculate the sun's true anomaly"""
    m = geom_mean_anomaly_sun(juliancentury)
    c = sun_eq_of_center(juliancentury)

    return m + c


def sun_rad_vector(juliancentury: float) -> float:
    v = sun_true_anomoly(juliancentury)
    e = eccentric_location_earth_orbit(juliancentury)

    return (1.000001018 * (1 - e * e)) / (1 + e * cos(radians(v)))


def sun_apparent_long(juliancentury: float) -> float:
    true_long = sun_true_long(juliancentury)

    omega = 125.04 - 1934.136 * juliancentury
    return true_long - 0.00569 - 0.00478 * sin(radians(omega))


def mean_obliquity_of_ecliptic(juliancentury: float) -> float:
    seconds = 21.448 - juliancentury * (
        46.815 + juliancentury * (0.00059 - juliancentury * (0.001813))
    )
    return 23.0 + (26.0 + (seconds / 60.0)) / 60.0


def obliquity_correction(juliancentury: float) -> float:
    e0 = mean_obliquity_of_ecliptic(juliancentury)

    omega = 125.04 - 1934.136 * juliancentury
    return e0 + 0.00256 * cos(radians(omega))


def sun_rt_ascension(juliancentury: float) -> float:
    """Calculate the sun's right ascension"""
    oc = obliquity_correction(juliancentury)
    al = sun_apparent_long(juliancentury)

    tananum = cos(radians(oc)) * sin(radians(al))
    tanadenom = cos(radians(al))

    return degrees(atan2(tananum, tanadenom))


def sun_declination(juliancentury: float) -> float:
    """Calculate the sun's declination"""
    e = obliquity_correction(juliancentury)
    lambd = sun_apparent_long(juliancentury)

    sint = sin(radians(e)) * sin(radians(lambd))
    return degrees(asin(sint))


def var_y(juliancentury: float) -> float:
    epsilon = obliquity_correction(juliancentury)
    y = tan(radians(epsilon) / 2.0)
    return y * y


def eq_of_time(juliancentury: float) -> Minutes:
    l0 = geom_mean_long_sun(juliancentury)
    e = eccentric_location_earth_orbit(juliancentury)
    m = geom_mean_anomaly_sun(juliancentury)

    y = var_y(juliancentury)

    sin2l0 = sin(2.0 * radians(l0))
    sinm = sin(radians(m))
    cos2l0 = cos(2.0 * radians(l0))
    sin4l0 = sin(4.0 * radians(l0))
    sin2m = sin(2.0 * radians(m))

    Etime = (
        y * sin2l0
        - 2.0 * e * sinm
        + 4.0 * e * y * sinm * cos2l0
        - 0.5 * y * y * sin4l0
        - 1.25 * e * e * sin2m
    )

    return degrees(Etime) * 4.0


def hour_angle(
    latitude: float, declination: float, zenith: float, direction: SunDirection
) -> float:
    """Calculate the hour angle of the sun

    See https://en.wikipedia.org/wiki/Hour_angle#Solar_hour_angle

    Args:
        latitude: The latitude of the obersver
        declination: The declination of the sun
        zenith: The zenith angle of the sun
        direction: The direction of traversal of the sun

    Raises:
        ValueError
    """

    latitude_rad = radians(latitude)
    declination_rad = radians(declination)
    zenith_rad = radians(zenith)

    h = (cos(zenith_rad) - sin(latitude_rad) * sin(declination_rad)) / (
        cos(latitude_rad) * cos(declination_rad)
    )

    hour_angle = acos(h)
    if direction == SunDirection.SETTING:
        hour_angle = -hour_angle
    return hour_angle


def adjust_to_horizon(elevation: float) -> float:
    """Calculate the extra degrees of depression that you can see round the earth
    due to the increase in elevation.

    Args:
        elevation: Elevation above the earth in metres

    Returns:
        A number of degrees to add to adjust for the elevation of the observer
    """

    if elevation <= 0:
        return 0

    r = 6356900  # radius of the earth
    a1 = r
    h1 = r + elevation
    theta1 = acos(a1 / h1)
    return degrees(theta1)


def adjust_to_obscuring_feature(elevation: Tuple[float, float]) -> float:
    """Calculate the number of degrees to adjust for an obscuring feature"""
    if elevation[0] == 0.0:
        return 0.0

    sign = -1 if elevation[0] < 0.0 else 1
    return sign * degrees(
        acos(fabs(elevation[0]) / sqrt(pow(elevation[0], 2) + pow(elevation[1], 2)))
    )


def time_of_transit(
    observer: Observer,
    date: datetime.date,
    zenith: float,
    direction: SunDirection,
    with_refraction: bool = True,
) -> datetime.datetime:
    """Calculate the time in the UTC timezone when the sun transits the
    specificed zenith

    Args:
        observer: An observer viewing the sun at a specific, latitude, longitude
            and elevation
        date: The date to calculate for
        zenith: The zenith angle for which to calculate the transit time
        direction: The direction that the sun is traversing

    Raises:
        ValueError if the zenith is not transitted by the sun

    Returns:
        the time when the sun transits the specificed zenith
    """
    if observer.latitude > 89.8:
        latitude = 89.8
    elif observer.latitude < -89.8:
        latitude = -89.8
    else:
        latitude = observer.latitude

    adjustment_for_elevation = 0.0
    if isinstance(observer.elevation, float) and observer.elevation > 0.0:
        adjustment_for_elevation = adjust_to_horizon(observer.elevation)
    elif isinstance(observer.elevation, tuple):
        adjustment_for_elevation = adjust_to_obscuring_feature(observer.elevation)

    if with_refraction:
        adjustment_for_refraction = refraction_at_zenith(
            zenith + adjustment_for_elevation
        )
    else:
        adjustment_for_refraction = 0.0

    jd = julianday(date)
    adjustment = 0.0
    timeUTC = 0.0

    for _ in range(2):
        jc = julianday_to_juliancentury(jd + adjustment)
        declination = sun_declination(jc)

        hourangle = hour_angle(
            latitude,
            declination,
            zenith + adjustment_for_elevation + adjustment_for_refraction,
            direction,
        )

        delta = -observer.longitude - degrees(hourangle)

        eqtime = eq_of_time(jc)
        offset = delta * 4.0 - eqtime

        if offset < -720.0:
            offset += 1440

        timeUTC = 720.0 + offset
        adjustment = timeUTC / 1440.0

    td = minutes_to_timedelta(timeUTC)
    dt = datetime.datetime(date.year, date.month, date.day) + td
    dt = dt.replace(tzinfo=datetime.timezone.utc)  # pylint: disable=E1120
    return dt


def time_at_elevation(
    observer: Observer,
    elevation: float,
    date: Optional[datetime.date] = None,
    direction: SunDirection = SunDirection.RISING,
    tzinfo: Union[str, datetime.tzinfo] = datetime.timezone.utc,
    with_refraction: bool = True,
) -> datetime.datetime:
    """Calculates the time when the sun is at the specified elevation on the
    specified date.

    Note:
        This method uses positive elevations for those above the horizon.

        Elevations greater than 90 degrees are converted to a setting sun
        i.e. an elevation of 110 will calculate a setting sun at 70 degrees.

    Args:
        elevation: Elevation of the sun in degrees above the horizon to calculate for.
        observer:  Observer to calculate for
        date:      Date to calculate for. Default is today's date in the timezone
                   `tzinfo`.
        direction: Determines whether the calculated time is for the sun rising
                   or setting.
                   Use ``SunDirection.RISING`` or ``SunDirection.SETTING``.
                   Default is rising.
        tzinfo:    Timezone to return times in. Default is UTC.

    Returns:
        Date and time at which the sun is at the specified elevation.
    """

    if elevation > 90.0:
        elevation = 180.0 - elevation
        direction = SunDirection.SETTING

    if isinstance(tzinfo, str):
        tzinfo = zoneinfo.ZoneInfo(tzinfo)  # type: ignore

    if date is None:
        date = today(tzinfo)  # type: ignore

    zenith = 90 - elevation
    try:
        return time_of_transit(
            observer, date, zenith, direction, with_refraction
        ).astimezone(
            tzinfo  # type: ignore
        )
    except ValueError as exc:
        if exc.args[0] == "math domain error":
            raise ValueError(
                f"Sun never reaches an elevation of {elevation} degrees "
                "at this location."
            ) from exc
        else:
            raise


def noon(
    observer: Observer,
    date: Optional[datetime.date] = None,
    tzinfo: Union[str, datetime.tzinfo] = datetime.timezone.utc,
) -> datetime.datetime:
    """Calculate solar noon time when the sun is at its highest point.

    Args:
        observer: An observer viewing the sun at a specific, latitude, longitude
                  and elevation
        date:     Date to calculate for. Default is today for the specified tzinfo.
        tzinfo:   Timezone to return times in. Default is UTC.

    Returns:
        Date and time at which noon occurs.
    """
    if isinstance(tzinfo, str):
        tzinfo = zoneinfo.ZoneInfo(tzinfo)  # type: ignore

    if date is None:
        date = today(tzinfo)  # type: ignore

    jc = julianday_to_juliancentury(julianday(date))
    eqtime = eq_of_time(jc)
    timeUTC = (720.0 - (4 * observer.longitude) - eqtime) / 60.0

    hour = int(timeUTC)
    minute = int((timeUTC - hour) * 60)
    second = int((((timeUTC - hour) * 60) - minute) * 60)

    if second > 59:
        second -= 60
        minute += 1
    elif second < 0:
        second += 60
        minute -= 1

    if minute > 59:
        minute -= 60
        hour += 1
    elif minute < 0:
        minute += 60
        hour -= 1

    if hour > 23:
        hour -= 24
        date += datetime.timedelta(days=1)
    elif hour < 0:
        hour += 24
        date -= datetime.timedelta(days=1)

    noon = datetime.datetime(
        date.year,
        date.month,
        date.day,
        hour,
        minute,
        second,
        tzinfo=datetime.timezone.utc,
    )
    return noon.astimezone(tzinfo)  # type: ignore pylint: disable=E1120


def midnight(
    observer: Observer,
    date: Optional[datetime.date] = None,
    tzinfo: Union[str, datetime.tzinfo] = datetime.timezone.utc,
) -> datetime.datetime:
    """Calculate solar midnight time.

    Note:
        This calculates the solar midnight that is closest
        to 00:00:00 of the specified date i.e. it may return a time that is on
        the previous day.

    Args:
        observer: An observer viewing the sun at a specific, latitude, longitude
                  and elevation
        date:     Date to calculate for. Default is today for the specified tzinfo.
        tzinfo:   Timezone to return times in. Default is UTC.

    Returns:
        Date and time at which midnight occurs.
    """
    if isinstance(tzinfo, str):
        tzinfo = zoneinfo.ZoneInfo(tzinfo)  # type: ignore

    if date is None:
        date = today(tzinfo)  # type: ignore

    midday = datetime.time(12, 0, 0)
    jd = julianday(datetime.datetime.combine(date, midday))
    newt = julianday_to_juliancentury(jd + 0.5 + -observer.longitude / 360.0)

    eqtime = eq_of_time(newt)
    timeUTC = (-observer.longitude * 4.0) - eqtime

    timeUTC = timeUTC / 60.0
    hour = int(timeUTC)
    minute = int((timeUTC - hour) * 60)
    second = int((((timeUTC - hour) * 60) - minute) * 60)

    if second > 59:
        second -= 60
        minute += 1
    elif second < 0:
        second += 60
        minute -= 1

    if minute > 59:
        minute -= 60
        hour += 1
    elif minute < 0:
        minute += 60
        hour -= 1

    if hour < 0:
        hour += 24
        date -= datetime.timedelta(days=1)

    midnight = datetime.datetime(
        date.year,
        date.month,
        date.day,
        hour,
        minute,
        second,
        tzinfo=datetime.timezone.utc,
    )
    return midnight.astimezone(tzinfo)  # type: ignore


def zenith_and_azimuth(
    observer: Observer,
    dateandtime: datetime.datetime,
    with_refraction: bool = True,
) -> Tuple[float, float]:
    if observer.latitude > 89.8:
        latitude = 89.8
    elif observer.latitude < -89.8:
        latitude = -89.8
    else:
        latitude = observer.latitude

    longitude = observer.longitude

    if dateandtime.tzinfo is None:
        zone = 0.0
        utc_datetime = dateandtime
    else:
        zone = -dateandtime.utcoffset().total_seconds() / 3600.0  # type: ignore
        utc_datetime = dateandtime.astimezone(datetime.timezone.utc)

    jd = julianday(utc_datetime)
    t = julianday_to_juliancentury(jd)
    declination = sun_declination(t)
    eqtime = eq_of_time(t)

    # 360deg * 4 == 1440 minutes, 60*24 = 1440 minutes == 1 rotation
    solarTimeFix = eqtime + (4.0 * longitude) + (60 * zone)
    trueSolarTime = (
        dateandtime.hour * 60.0
        + dateandtime.minute
        + dateandtime.second / 60.0
        + solarTimeFix
    )
    #    in minutes as a float, fractional part is seconds

    while trueSolarTime > 1440:
        trueSolarTime = trueSolarTime - 1440

    hourangle = trueSolarTime / 4.0 - 180.0
    #    Thanks to Louis Schwarzmayr for the next line:
    if hourangle < -180:
        hourangle = hourangle + 360.0

    ch = cos(radians(hourangle))
    # sh = sin(radians(hourangle))
    cl = cos(radians(latitude))
    sl = sin(radians(latitude))
    sd = sin(radians(declination))
    cd = cos(radians(declination))

    csz = cl * cd * ch + sl * sd

    if csz > 1.0:
        csz = 1.0
    elif csz < -1.0:
        csz = -1.0

    zenith = degrees(acos(csz))

    azDenom = cl * sin(radians(zenith))

    if abs(azDenom) > 0.001:
        azRad = ((sl * cos(radians(zenith))) - sd) / azDenom

        if abs(azRad) > 1.0:
            if azRad < 0:
                azRad = -1.0
            else:
                azRad = 1.0

        azimuth = 180.0 - degrees(acos(azRad))

        if hourangle > 0.0:
            azimuth = -azimuth
    else:
        if latitude > 0.0:
            azimuth = 180.0
        else:
            azimuth = 0.0

    if azimuth < 0.0:
        azimuth = azimuth + 360.0

    if with_refraction:
        zenith -= refraction_at_zenith(zenith)
        # elevation = 90 - zenith

    return zenith, azimuth


def zenith(
    observer: Observer,
    dateandtime: Optional[datetime.datetime] = None,
    with_refraction: bool = True,
) -> float:
    """Calculate the zenith angle of the sun.

    Args:
        observer:    Observer to calculate the solar zenith for
        dateandtime: The date and time for which to calculate the angle.
                     If `dateandtime` is None or is a naive Python datetime
                     then it is assumed to be in the UTC timezone.
        with_refraction: If True adjust zenith to take refraction into account

    Returns:
        The zenith angle in degrees.
    """

    if dateandtime is None:
        dateandtime = now(datetime.timezone.utc)

    return zenith_and_azimuth(observer, dateandtime, with_refraction)[0]


def azimuth(
    observer: Observer,
    dateandtime: Optional[datetime.datetime] = None,
) -> float:
    """Calculate the azimuth angle of the sun.

    Args:
        observer:    Observer to calculate the solar azimuth for
        dateandtime: The date and time for which to calculate the angle.
                     If `dateandtime` is None or is a naive Python datetime
                     then it is assumed to be in the UTC timezone.

    Returns:
        The azimuth angle in degrees clockwise from North.

    If `dateandtime` is a naive Python datetime then it is assumed to be
    in the UTC timezone.
    """

    if dateandtime is None:
        dateandtime = now(datetime.timezone.utc)

    return zenith_and_azimuth(observer, dateandtime)[1]


def elevation(
    observer: Observer,
    dateandtime: Optional[datetime.datetime] = None,
    with_refraction: bool = True,
) -> float:
    """Calculate the sun's angle of elevation.

    Args:
        observer:    Observer to calculate the solar elevation for
        dateandtime: The date and time for which to calculate the angle.
                     If `dateandtime` is None or is a naive Python datetime
                     then it is assumed to be in the UTC timezone.
        with_refraction: If True adjust elevation to take refraction into account

    Returns:
        The elevation angle in degrees above the horizon.
    """

    if dateandtime is None:
        dateandtime = now(datetime.timezone.utc)

    return 90.0 - zenith(observer, dateandtime, with_refraction)


def dawn(
    observer: Observer,
    date: Optional[datetime.date] = None,
    depression: Union[float, Depression] = Depression.CIVIL,
    tzinfo: Union[str, datetime.tzinfo] = datetime.timezone.utc,
) -> datetime.datetime:
    """Calculate dawn time.

    Args:
        observer:   Observer to calculate dawn for
        date:       Date to calculate for. Default is today's date in the
                    timezone `tzinfo`.
        depression: Number of degrees below the horizon to use to calculate dawn.
                    Default is for Civil dawn i.e. 6.0
        tzinfo:     Timezone to return times in. Default is UTC.

    Returns:
        Date and time at which dawn occurs.

    Raises:
        ValueError: if dawn does not occur on the specified date
    """
    if isinstance(tzinfo, str):
        tzinfo = zoneinfo.ZoneInfo(tzinfo)  # type: ignore

    if date is None:
        date = today(tzinfo)  # type: ignore
    elif isinstance(date, datetime.datetime):
        tzinfo = date.tzinfo or tzinfo
        date = date.date()

    dep: float = 0.0
    if isinstance(depression, Depression):
        dep = depression.value
    else:
        dep = depression

    try:
        tot = time_of_transit(
            observer, date, 90.0 + dep, SunDirection.RISING
        ).astimezone(
            tzinfo  # type: ignore
        )

        # If the dates don't match search on either the next or previous day.
        tot_date = tot.date()
        if tot_date != date:
            if tot_date < date:
                delta = datetime.timedelta(days=1)
            else:
                delta = datetime.timedelta(days=-1)
            new_date = date + delta

            tot = time_of_transit(
                observer,
                new_date,
                90.0 + dep,
                SunDirection.RISING,
            ).astimezone(
                tzinfo  # type: ignore
            )
            # Still can't get a time then raise the error
            tot_date = tot.date()
            if tot_date != date:
                raise ValueError("Unable to find a dawn time on the date specified")
        return tot
    except ValueError as exc:
        if exc.args[0] == "math domain error":
            raise ValueError(
                f"Sun never reaches {dep} degrees below the horizon, at this location."
            ) from exc
        else:
            raise


def sunrise(
    observer: Observer,
    date: Optional[datetime.date] = None,
    tzinfo: Union[str, datetime.tzinfo] = datetime.timezone.utc,
) -> datetime.datetime:
    """Calculate sunrise time.

    Args:
        observer: Observer to calculate sunrise for
        date:     Date to calculate for. Default is today's date in the
                  timezone `tzinfo`.
        tzinfo:   Timezone to return times in. Default is UTC.

    Returns:
        Date and time at which sunrise occurs.

    Raises:
        ValueError: if the sun does not reach the horizon on the specified date
    """
    if isinstance(tzinfo, str):
        tzinfo = zoneinfo.ZoneInfo(tzinfo)  # type: ignore

    if date is None:
        date = today(tzinfo)  # type: ignore
    elif isinstance(date, datetime.datetime):
        tzinfo = date.tzinfo or tzinfo
        date = date.date()

    try:
        tot = time_of_transit(
            observer,
            date,
            90.0 + SUN_APPARENT_RADIUS,
            SunDirection.RISING,
        ).astimezone(
            tzinfo  # type: ignore
        )

        tot_date = tot.date()
        if tot_date != date:
            if tot_date < date:
                delta = datetime.timedelta(days=1)
            else:
                delta = datetime.timedelta(days=-1)
            new_date = date + delta

            tot = time_of_transit(
                observer,
                new_date,
                90.0 + SUN_APPARENT_RADIUS,
                SunDirection.RISING,
            ).astimezone(
                tzinfo  # type: ignore
            )
            tot_date = tot.date()
            if tot_date != date:
                raise ValueError("Unable to find a sunrise time on the date specified")
        return tot
    except ValueError as exc:
        if exc.args[0] == "math domain error":
            z = zenith(observer, noon(observer, date))
            if z > 90.0:
                msg = "Sun is always below the horizon on this day, at this location."
            else:
                msg = "Sun is always above the horizon on this day, at this location."
            raise ValueError(msg) from exc
        else:
            raise


def sunset(
    observer: Observer,
    date: Optional[datetime.date] = None,
    tzinfo: Union[str, datetime.tzinfo] = datetime.timezone.utc,
) -> datetime.datetime:
    """Calculate sunset time.

    Args:
        observer: Observer to calculate sunset for
        date:     Date to calculate for. Default is today's date in the
                  timezone `tzinfo`.
        tzinfo:   Timezone to return times in. Default is UTC.

    Returns:
        Date and time at which sunset occurs.

    Raises:
        ValueError: if the sun does not reach the horizon
    """

    if isinstance(tzinfo, str):
        tzinfo = zoneinfo.ZoneInfo(tzinfo)  # type: ignore

    if date is None:
        date = today(tzinfo)  # type: ignore
    elif isinstance(date, datetime.datetime):
        tzinfo = date.tzinfo or tzinfo
        date = date.date()

    try:
        tot = time_of_transit(
            observer,
            date,
            90.0 + SUN_APPARENT_RADIUS,
            SunDirection.SETTING,
        ).astimezone(
            tzinfo  # type: ignore
        )

        tot_date = tot.date()
        if tot_date != date:
            if tot_date < date:
                delta = datetime.timedelta(days=1)
            else:
                delta = datetime.timedelta(days=-1)
            new_date = date + delta

            tot = time_of_transit(
                observer,
                new_date,
                90.0 + SUN_APPARENT_RADIUS,
                SunDirection.SETTING,
            ).astimezone(
                tzinfo  # type: ignore
            )
            tot_date = tot.date()
            if tot_date != date:
                raise ValueError("Unable to find a sunset time on the date specified")
        return tot
    except ValueError as exc:
        if exc.args[0] == "math domain error":
            z = zenith(observer, noon(observer, date))
            if z > 90.0:
                msg = "Sun is always below the horizon on this day, at this location."
            else:
                msg = "Sun is always above the horizon on this day, at this location."
            raise ValueError(msg) from exc
        else:
            raise


def dusk(
    observer: Observer,
    date: Optional[datetime.date] = None,
    depression: Union[float, Depression] = Depression.CIVIL,
    tzinfo: Union[str, datetime.tzinfo] = datetime.timezone.utc,
) -> datetime.datetime:
    """Calculate dusk time.

    Args:
        observer:   Observer to calculate dusk for
        date:       Date to calculate for. Default is today's date in the
                    timezone `tzinfo`.
        depression: Number of degrees below the horizon to use to calculate dusk.
                    Default is for Civil dusk i.e. 6.0
        tzinfo:     Timezone to return times in. Default is UTC.

    Returns:
        Date and time at which dusk occurs.

    Raises:
        ValueError: if dusk does not occur on the specified date
    """

    if isinstance(tzinfo, str):
        tzinfo = zoneinfo.ZoneInfo(tzinfo)  # type: ignore

    if date is None:
        date = today(tzinfo)  # type: ignore
    elif isinstance(date, datetime.datetime):
        tzinfo = date.tzinfo or tzinfo
        date = date.date()

    dep: float = 0.0
    if isinstance(depression, Depression):
        dep = depression.value
    else:
        dep = depression

    try:
        tot = time_of_transit(
            observer, date, 90.0 + dep, SunDirection.SETTING
        ).astimezone(
            tzinfo  # type: ignore
        )

        tot_date = tot.date()
        if tot_date != date:
            if tot_date < date:
                delta = datetime.timedelta(days=1)
            else:
                delta = datetime.timedelta(days=-1)
            new_date = date + delta

            tot = time_of_transit(
                observer,
                new_date,
                90.0 + dep,
                SunDirection.SETTING,
            ).astimezone(
                tzinfo  # type: ignore
            )
            tot_date = tot.date()
            if tot_date != date:
                raise ValueError("Unable to find a dusk time on the date specified")
        return tot
    except ValueError as exc:
        if exc.args[0] == "math domain error":
            raise ValueError(
                f"Sun never reaches {dep} degrees below the horizon, at this location."
            ) from exc
        else:
            raise


def daylight(
    observer: Observer,
    date: Optional[datetime.date] = None,
    tzinfo: Union[str, datetime.tzinfo] = datetime.timezone.utc,
) -> TimePeriod:
    """Calculate daylight start and end times.

    Args:
        observer:   Observer to calculate daylight for
        date:       Date to calculate for. Default is today's date in the
                    timezone `tzinfo`.
        tzinfo:     Timezone to return times in. Default is UTC.

    Returns:
        A tuple of the date and time at which daylight starts and ends.

    Raises:
        ValueError: if the sun does not rise or does not set
    """
    if isinstance(tzinfo, str):
        tzinfo = zoneinfo.ZoneInfo(tzinfo)  # type: ignore

    if date is None:
        date = today(tzinfo)  # type: ignore

    sr = sunrise(observer, date, tzinfo)
    ss = sunset(observer, date, tzinfo)

    return sr, ss


def night(
    observer: Observer,
    date: Optional[datetime.date] = None,
    tzinfo: Union[str, datetime.tzinfo] = datetime.timezone.utc,
) -> TimePeriod:
    """Calculate night start and end times.

    Night is calculated to be between astronomical dusk on the
    date specified and astronomical dawn of the next day.

    Args:
        observer:   Observer to calculate night for
        date:       Date to calculate for. Default is today's date for the
                    specified tzinfo.
        tzinfo:     Timezone to return times in. Default is UTC.

    Returns:
        A tuple of the date and time at which night starts and ends.

    Raises:
        ValueError: if dawn does not occur on the specified date or
                    dusk on the following day
    """
    if isinstance(tzinfo, str):
        tzinfo = zoneinfo.ZoneInfo(tzinfo)  # type: ignore

    if date is None:
        date = today(tzinfo)  # type: ignore

    start = dusk(observer, date, 6, tzinfo)
    tomorrow = date + datetime.timedelta(days=1)
    end = dawn(observer, tomorrow, 6, tzinfo)

    return start, end


def twilight(
    observer: Observer,
    date: Optional[datetime.date] = None,
    direction: SunDirection = SunDirection.RISING,
    tzinfo: Union[str, datetime.tzinfo] = datetime.timezone.utc,
) -> TimePeriod:
    """Returns the start and end times of Twilight
    when the sun is traversing in the specified direction.

    This method defines twilight as being between the time
    when the sun is at -6 degrees and sunrise/sunset.

    Args:
        observer:   Observer to calculate twilight for
        date:       Date for which to calculate the times.
                    Default is today's date in the timezone `tzinfo`.
        direction:  Determines whether the time is for the sun rising or setting.
                    Use ``astral.SunDirection.RISING`` or
                    ``astral.SunDirection.SETTING``.
        tzinfo:     Timezone to return times in. Default is UTC.

    Returns:
        A tuple of the date and time at which twilight starts and ends.

    Raises:
        ValueError: if the sun does not rise or does not set
    """

    if isinstance(tzinfo, str):
        tzinfo = zoneinfo.ZoneInfo(tzinfo)  # type: ignore

    if date is None:
        date = today(tzinfo)  # type: ignore

    start = time_of_transit(observer, date, 90 + 6, direction,).astimezone(
        tzinfo  # type: ignore
    )
    if direction == SunDirection.RISING:
        end = sunrise(observer, date, tzinfo).astimezone(tzinfo)  # type: ignore
    else:
        end = sunset(observer, date, tzinfo).astimezone(tzinfo)  # type: ignore

    if direction == SunDirection.RISING:
        return start, end
    else:
        return end, start


def golden_hour(
    observer: Observer,
    date: Optional[datetime.date] = None,
    direction: SunDirection = SunDirection.RISING,
    tzinfo: Union[str, datetime.tzinfo] = datetime.timezone.utc,
) -> TimePeriod:
    """Returns the start and end times of the Golden Hour
    when the sun is traversing in the specified direction.

    This method uses the definition from PhotoPills i.e. the
    golden hour is when the sun is between 4 degrees below the horizon
    and 6 degrees above.

    Args:
        observer:   Observer to calculate the golden hour for
        date:       Date for which to calculate the times.
                    Default is today's date in the timezone `tzinfo`.
        direction:  Determines whether the time is for the sun rising or setting.
                    Use ``SunDirection.RISING`` or ``SunDirection.SETTING``.
        tzinfo:     Timezone to return times in. Default is UTC.

    Returns:
        A tuple of the date and time at which the Golden Hour starts and ends.

    Raises:
        ValueError: if the sun does not transit the elevations -4 & +6 degrees
    """

    if isinstance(tzinfo, str):
        tzinfo = zoneinfo.ZoneInfo(tzinfo)  # type: ignore

    if date is None:
        date = today(tzinfo)  # type: ignore

    start = time_of_transit(observer, date, 90 + 4, direction,).astimezone(
        tzinfo  # type: ignore
    )
    end = time_of_transit(observer, date, 90 - 6, direction,).astimezone(
        tzinfo  # type: ignore
    )

    if direction == SunDirection.RISING:
        return start, end
    else:
        return end, start


def blue_hour(
    observer: Observer,
    date: Optional[datetime.date] = None,
    direction: SunDirection = SunDirection.RISING,
    tzinfo: Union[str, datetime.tzinfo] = datetime.timezone.utc,
) -> TimePeriod:
    """Returns the start and end times of the Blue Hour
    when the sun is traversing in the specified direction.

    This method uses the definition from PhotoPills i.e. the
    blue hour is when the sun is between 6 and 4 degrees below the horizon.

    Args:
        observer:   Observer to calculate the blue hour for
        date:       Date for which to calculate the times.
                    Default is today's date in the timezone `tzinfo`.
        direction:  Determines whether the time is for the sun rising or setting.
                    Use ``SunDirection.RISING`` or ``SunDirection.SETTING``.
        tzinfo:     Timezone to return times in. Default is UTC.

    Returns:
        A tuple of the date and time at which the Blue Hour starts and ends.

    Raises:
        ValueError: if the sun does not transit the elevations -4 & -6 degrees
    """

    if isinstance(tzinfo, str):
        tzinfo = zoneinfo.ZoneInfo(tzinfo)  # type: ignore

    if date is None:
        date = today(tzinfo)  # type: ignore

    start = time_of_transit(observer, date, 90 + 6, direction,).astimezone(
        tzinfo  # type: ignore
    )
    end = time_of_transit(observer, date, 90 + 4, direction,).astimezone(
        tzinfo  # type: ignore
    )

    if direction == SunDirection.RISING:
        return start, end
    else:
        return end, start


def rahukaalam(
    observer: Observer,
    date: Optional[datetime.date] = None,
    daytime: bool = True,
    tzinfo: Union[str, datetime.tzinfo] = datetime.timezone.utc,
) -> TimePeriod:
    """Calculate ruhakaalam times.

    Args:
        observer:   Observer to calculate rahukaalam for
        date:       Date to calculate for. Default is today's date in the
                    timezone `tzinfo`.
        daytime:    If True calculate for the day time else calculate for the
                    night time.
        tzinfo:     Timezone to return times in. Default is UTC.

    Returns:
        Tuple containing the start and end times for Rahukaalam.

    Raises:
        ValueError: if the sun does not rise or does not set
    """

    if isinstance(tzinfo, str):
        tzinfo = zoneinfo.ZoneInfo(tzinfo)  # type: ignore

    if date is None:
        date = today(tzinfo)  # type: ignore

    if daytime:
        start = sunrise(observer, date, tzinfo)
        end = sunset(observer, date, tzinfo)
    else:
        start = sunset(observer, date, tzinfo)
        oneday = datetime.timedelta(days=1)
        end = sunrise(observer, date + oneday, tzinfo)

    octant_duration = datetime.timedelta(seconds=(end - start).seconds / 8)

    # Mo,Sa,Fr,We,Th,Tu,Su
    octant_index = [1, 6, 4, 5, 3, 2, 7]

    weekday = date.weekday()
    octant = octant_index[weekday]

    start = start + (octant_duration * octant)
    end = start + octant_duration

    return start, end


def sun(
    observer: Observer,
    date: Optional[datetime.date] = None,
    dawn_dusk_depression: Union[float, Depression] = Depression.CIVIL,
    tzinfo: Union[str, datetime.tzinfo] = datetime.timezone.utc,
) -> Dict[str, datetime.datetime]:
    """Calculate all the info for the sun at once.

    Args:
        observer:             Observer for which to calculate the times of the sun
        date:                 Date to calculate for.
                              Default is today's date in the timezone `tzinfo`.
        dawn_dusk_depression: Depression to use to calculate dawn and dusk.
                              Default is for Civil dusk i.e. 6.0
        tzinfo:               Timezone to return times in. Default is UTC.

    Returns:
        Dictionary with keys ``dawn``, ``sunrise``, ``noon``, ``sunset`` and ``dusk``
        whose values are the results of the corresponding functions.

    Raises:
        ValueError: if passed through from any of the functions
    """

    if isinstance(tzinfo, str):
        tzinfo = zoneinfo.ZoneInfo(tzinfo)  # type: ignore

    if date is None:
        date = today(tzinfo)  # type: ignore

    return {
        "dawn": dawn(observer, date, dawn_dusk_depression, tzinfo),
        "sunrise": sunrise(observer, date, tzinfo),
        "noon": noon(observer, date, tzinfo),
        "sunset": sunset(observer, date, tzinfo),
        "dusk": dusk(observer, date, dawn_dusk_depression, tzinfo),
    }



import pytest

from astral import Observer


class TestObserver:
    def test_default(self):
        obs = Observer()
        assert obs.latitude == 51.4733
        assert obs.longitude == -0.0008333
        assert obs.elevation == 0.0

    def test_from_float(self):
        obs = Observer(1, 1, 1)
        assert obs.latitude == 1.0
        assert obs.longitude == 1.0
        assert obs.elevation == 1.0

    def test_from_string(self):
        obs = Observer("1", "2", "3")
        assert obs.latitude == 1.0
        assert obs.longitude == 2.0
        assert obs.elevation == 3.0

    def test_from_dms(self):
        obs = Observer("24°N", "22°30'S", "3")
        assert obs.latitude == 24.0
        assert obs.longitude == -22.5
        assert obs.elevation == 3.0

    def test_bad_latitude(self):
        with pytest.raises(ValueError):
            Observer("o", 1, 1)

    def test_bad_longitude(self):
        with pytest.raises(ValueError):
            Observer(1, "o", 1)

    def test_bad_elevation(self):
        with pytest.raises(ValueError):
            Observer(1, 1, "o")

    def test_latitude_outside_limits(self):
        obs = Observer(90.1, 0, 0)
        assert obs.latitude == 90.0
        obs = Observer(-90.1, 0, 0)
        assert obs.latitude == -90.0

    def test_longitude_outside_limits(self):
        obs = Observer(0, 180.1, 0)
        assert obs.longitude == 180.0
        obs = Observer(0, -180.1, 0)
        assert obs.longitude == -180.0



# -*- coding: utf-8 -*-
# Test data taken from http://www.timeanddate.com/sun/uk/london

import datetime

import freezegun
import pytest

from almost_equal import datetime_almost_equal
from astral import TimePeriod, sun
from astral.location import Location
from astral.sun import SunDirection


class TestGoldenHour:
    """Tests for the golden_hour function"""

    @pytest.mark.parametrize(
        "day,golden_hour",
        [
            (
                datetime.date(2015, 12, 1),
                (
                    datetime.datetime(2015, 12, 1, 1, 10, 10),
                    datetime.datetime(2015, 12, 1, 2, 0, 43),
                ),
            ),
            (
                datetime.date(2016, 1, 1),
                (
                    datetime.datetime(2016, 1, 1, 1, 27, 46),
                    datetime.datetime(2016, 1, 1, 2, 19, 1),
                ),
            ),
        ],
    )
    def test_morning(
        self, day: datetime.date, golden_hour: TimePeriod, new_delhi: Location
    ):
        start1 = golden_hour[0].replace(tzinfo=datetime.timezone.utc)
        end1 = golden_hour[1].replace(tzinfo=datetime.timezone.utc)

        start2, end2 = sun.golden_hour(new_delhi.observer, day, SunDirection.RISING)
        assert datetime_almost_equal(end1, end2, seconds=90)
        assert datetime_almost_equal(start1, start2, seconds=90)

    def test_evening(self, london: Location):
        test_data = {
            datetime.date(2016, 5, 18): (
                datetime.datetime(2016, 5, 18, 19, 2),
                datetime.datetime(2016, 5, 18, 20, 17),
            )
        }

        for day, golden_hour in test_data.items():
            start1 = golden_hour[0].replace(tzinfo=datetime.timezone.utc)
            end1 = golden_hour[1].replace(tzinfo=datetime.timezone.utc)

            start2, end2 = sun.golden_hour(london.observer, day, SunDirection.SETTING)
            assert datetime_almost_equal(end1, end2, seconds=60)
            assert datetime_almost_equal(start1, start2, seconds=60)

    @freezegun.freeze_time("2015-12-1")
    def test_no_date(self, new_delhi: Location):
        start = datetime.datetime(2015, 12, 1, 1, 10, 10, tzinfo=datetime.timezone.utc)
        end = datetime.datetime(2015, 12, 1, 2, 0, 43, tzinfo=datetime.timezone.utc)
        ans = sun.golden_hour(new_delhi.observer)
        assert datetime_almost_equal(ans[0], start, 90)
        assert datetime_almost_equal(ans[1], end, 90)


class TestBlueHour:
    """Tests for the blue_hour function"""

    def test_morning(self, london: Location):
        test_data = {
            datetime.date(2016, 5, 19): (
                datetime.datetime(2016, 5, 19, 3, 19),
                datetime.datetime(2016, 5, 19, 3, 36),
            )
        }

        for day, blue_hour in test_data.items():
            start1 = blue_hour[0].replace(tzinfo=datetime.timezone.utc)
            end1 = blue_hour[1].replace(tzinfo=datetime.timezone.utc)

            start2, end2 = sun.blue_hour(london.observer, day, SunDirection.RISING)
            assert datetime_almost_equal(end1, end2, seconds=90)
            assert datetime_almost_equal(start1, start2, seconds=90)

    def test_evening(self, london: Location):
        test_data = {
            datetime.date(2016, 5, 19): (
                datetime.datetime(2016, 5, 19, 20, 18),
                datetime.datetime(2016, 5, 19, 20, 35),
            )
        }

        for day, blue_hour in test_data.items():
            start1 = blue_hour[0].replace(tzinfo=datetime.timezone.utc)
            end1 = blue_hour[1].replace(tzinfo=datetime.timezone.utc)

            start2, end2 = sun.blue_hour(london.observer, day, SunDirection.SETTING)
            assert datetime_almost_equal(end1, end2, seconds=90)
            assert datetime_almost_equal(start1, start2, seconds=90)

    @freezegun.freeze_time("2016-5-19")
    def test_no_date(self, london: Location):
        start = datetime.datetime(2016, 5, 19, 20, 18, tzinfo=datetime.timezone.utc)
        end = datetime.datetime(2016, 5, 19, 20, 35, tzinfo=datetime.timezone.utc)
        ans = sun.blue_hour(london.observer, direction=SunDirection.SETTING)
        assert datetime_almost_equal(ans[0], start, 90)
        assert datetime_almost_equal(ans[1], end, 90)



import pytest
import astral
from astral.location import Location
from datetime import datetime, timedelta, timezone
from astral import sun


def _next_event(obs: astral.Observer, dt: datetime, event: str):
    for offset in range(0, 365):
        newdate = dt + timedelta(days=offset)
        try:
            t = getattr(sun, event)(date=newdate, observer=obs)
            return t
        except ValueError:
            pass
    assert False, "Should be unreachable"  # pragma: no cover


def test_NorwaySunUp(tromso: Location):
    """Test location in Norway where the sun doesn't set in summer."""
    june = datetime(2019, 6, 5, tzinfo=timezone.utc)

    with pytest.raises(ValueError):
        sun.sunrise(tromso.observer, june)
    with pytest.raises(ValueError):
        sun.sunset(tromso.observer, june)

    # Find the next sunset and sunrise:
    next_sunrise = _next_event(tromso.observer, june, "sunrise")
    next_sunset = _next_event(tromso.observer, june, "sunset")

    assert next_sunset < next_sunrise



"""Moon rise and set times

Right ascension, declination and distance of moon calcaulation
from

LOW-PRECISION FORMULAE FOR PLANETARY POSITIONS
http://articles.adsabs.harvard.edu/pdf/1979ApJS...41..391V
"""

import datetime
from dataclasses import dataclass, field, replace
from math import asin, atan2, cos, degrees, fabs, pi, radians, sin, sqrt
from typing import Callable, List, Optional, Union

try:
    import zoneinfo
except ImportError:
    from backports import zoneinfo

from astral import AstralBodyPosition, Observer, now, today
from astral.julian import julianday, julianday_2000
from astral.sidereal import lmst
from astral.table4 import Table4Row, table4_u, table4_v, table4_w

__all__ = ["phase"]

# Using 1896 arc seconds as moon's apparent diameter
MOON_APPARENT_RADIUS = 1896.0 / (60.0 * 60.0)

Degrees = float
Radians = float
Revolutions = float
ArgumentFunc = Optional[Callable[[float], float]]


@dataclass
class NoTransit:
    parallax: float = field(default_factory=float)


@dataclass
class TransitEvent:
    event: str
    when: datetime.time = field(default_factory=datetime.time)
    azimuth: float = field(default_factory=float)
    distance: float = field(default_factory=float)


def interpolate(f0: float, f1: float, f2: float, p: float) -> float:
    """3-point interpolation"""
    a = f1 - f0
    b = f2 - f1 - a
    f = f0 + p * (2 * a + b * (2 * p - 1))
    return f


def sgn(value1: Union[float, datetime.timedelta]) -> int:
    """Test whether value1 and value2 have the same sign"""
    if isinstance(value1, datetime.timedelta):
        value1 = value1.total_seconds()

    if value1 < 0:
        return -1
    elif value1 > 0:
        return 1
    else:
        return 0


def moon_mean_longitude(jd2000: float) -> Revolutions:
    _mean_longitude = 0.606434 + 0.03660110129 * jd2000
    _mean_longitude = _mean_longitude - int(_mean_longitude)
    return _mean_longitude


def moon_mean_anomoly(jd2000: float) -> Revolutions:
    _mean_anomoly = 0.374897 + 0.03629164709 * jd2000
    _mean_anomoly = _mean_anomoly - int(_mean_anomoly)
    return _mean_anomoly


def moon_argument_of_latitude(jd2000: float) -> Revolutions:
    _argument_of_latitude = 0.259091 + 0.03674819520 * jd2000
    _argument_of_latitude = _argument_of_latitude - int(_argument_of_latitude)
    return _argument_of_latitude


def moon_mean_elongation_from_sun(jd2000: float) -> Revolutions:
    _mean_elongation_from_sun = 0.827362 + 0.03386319198 * jd2000
    _mean_elongation_from_sun = _mean_elongation_from_sun - int(
        _mean_elongation_from_sun
    )
    return _mean_elongation_from_sun


def longitude_lunar_ascending_node(jd2000: float) -> Revolutions:
    _longitude_lunar_ascending_node = moon_mean_longitude(
        jd2000
    ) - moon_argument_of_latitude(jd2000)
    return _longitude_lunar_ascending_node


def sun_mean_longitude(jd2000: float) -> Revolutions:
    _sun_mean_longitude = 0.779072 + 0.00273790931 * jd2000
    _sun_mean_longitude = _sun_mean_longitude - int(_sun_mean_longitude)
    return _sun_mean_longitude


def sun_mean_anomoly(jd2000: float) -> Revolutions:
    _sun_mean_anomoly = 0.993126 + 0.00273777850 * jd2000
    _sun_mean_anomoly = _sun_mean_anomoly - int(_sun_mean_anomoly)
    return _sun_mean_anomoly


def venus_mean_longitude(jd2000: float) -> Revolutions:
    _venus_mean_longitude = 0.505498 + 0.00445046867 * jd2000
    _venus_mean_longitude = _venus_mean_longitude - int(_venus_mean_longitude)
    return _venus_mean_longitude


def moon_position(jd2000: float) -> AstralBodyPosition:
    """Calculate right ascension, declination and geocentric distance for the moon"""

    argument_values: List[Union[float, None]] = [
        moon_mean_longitude(jd2000),  # 1 = Lm
        moon_mean_anomoly(jd2000),  # 2 = Gm
        moon_argument_of_latitude(jd2000),  # 3 = Fm
        moon_mean_elongation_from_sun(jd2000),  # 4 = D
        longitude_lunar_ascending_node(jd2000),  # 5 = Om
        None,  # 6
        sun_mean_longitude(jd2000),  # 7 = Ls
        sun_mean_anomoly(jd2000),  # 8 = Gs
        None,  # 9
        None,  # 10
        None,  # 11
        venus_mean_longitude(jd2000),  # 12 = L2
    ]

    T = jd2000 / 36525 + 1

    def _calc_value(table: List[Table4Row]) -> float:
        result = 0.0
        for row in table:
            revolutions: float = 0.0
            for arg_number, multiplier in row.argument_multiplers.items():
                if multiplier != 0:
                    arg_value = argument_values[arg_number - 1]
                    if arg_value:
                        value = arg_value * multiplier
                        revolutions += value
                    else:
                        raise ValueError

            t_multipler = T if row.t else 1
            result += row.coefficient * t_multipler * row.sincos(revolutions * 2 * pi)

        return result

    v = _calc_value(table4_v)
    u = _calc_value(table4_u)
    w = _calc_value(table4_w)

    s = w / sqrt(u - v * v)
    right_ascension = asin(s) + (argument_values[0] or 0) * 2 * pi  # In radians

    s = v / sqrt(u)
    declination = asin(s)  # In radians

    distance = 60.40974 * sqrt(u)  # In Earth radii (≈6378km)

    return AstralBodyPosition(right_ascension, declination, distance)


def moon_transit_event(
    hour: float,
    lmst: Degrees,
    latitude: Degrees,
    distance: float,
    window: List[AstralBodyPosition],
) -> Union[TransitEvent, NoTransit]:
    """Check if the moon transits the horizon within the window.

    Args:
        hour: Hour of the day
        lmst: Local mean sidereal time in degrees
        latitude: Observer latitude
        distance: Distance to the moon
        window: Sliding window of moon positions that covers a part of the day
    """
    mst = radians(lmst)
    hour_angle = [0.0, 0.0, 0.0]

    k1 = radians(15 * 1.0027379097096138907193594760917)

    if window[2].right_ascension < window[0].right_ascension:
        window[2].right_ascension = window[2].right_ascension + 2 * pi

    hour_angle[0] = mst - window[0].right_ascension + (hour * k1)
    hour_angle[2] = mst - window[2].right_ascension + (hour * k1) + k1
    hour_angle[1] = (hour_angle[2] + hour_angle[0]) / 2

    window[1].declination = (window[2].declination + window[0].declination) / 2

    sl = sin(radians(latitude))
    cl = cos(radians(latitude))

    # moon apparent radius + parallax correction
    z = cos(radians(90 + MOON_APPARENT_RADIUS - (41.685 / distance)))

    if hour == 0:
        window[0].distance = (
            sl * sin(window[0].declination)
            + cl * cos(window[0].declination) * cos(hour_angle[0])
            - z
        )

    window[2].distance = (
        sl * sin(window[2].declination)
        + cl * cos(window[2].declination) * cos(hour_angle[2])
        - z
    )

    if sgn(window[0].distance) == sgn(window[2].distance):
        return NoTransit(window[2].distance)

    window[1].distance = (
        sl * sin(window[1].declination)
        + cl * cos(window[1].declination) * cos(hour_angle[1])
        - z
    )

    a = 2 * window[2].distance - 4 * window[1].distance + 2 * window[0].distance
    b = 4 * window[1].distance - 3 * window[0].distance - window[2].distance
    discriminant = b * b - 4 * a * window[0].distance

    if discriminant < 0:
        return NoTransit(window[2].distance)

    discriminant = sqrt(discriminant)
    e = (-b + discriminant) / (2 * a)
    if e > 1 or e < 0:
        e = (-b - discriminant) / (2 * a)

    time = hour + e + 1 / 120

    h = int(time)
    m = int((time - h) * 60)

    sd = sin(window[1].declination)
    cd = cos(window[1].declination)

    hour_angle_crossing = hour_angle[0] + e * (hour_angle[2] - hour_angle[0])
    sh = sin(hour_angle_crossing)
    ch = cos(hour_angle_crossing)

    x = cl * sd - sl * cd * ch
    y = -cd * sh

    az = degrees(atan2(y, x))
    if az < 0:
        az += 360
    if az > 360:
        az -= 360

    event_time = datetime.time(h, m, 0)
    if window[0].distance < 0 and window[2].distance > 0:
        return TransitEvent("rise", event_time, az, window[2].distance)

    if window[0].distance > 0 and window[2].distance < 0:
        return TransitEvent("set", event_time, az, window[2].distance)

    return NoTransit(window[2].distance)


def riseset(
    on: datetime.date,
    observer: Observer,
):
    """Calculate rise and set times"""
    jd2000 = julianday_2000(on)
    t0 = lmst(
        on,
        observer.longitude,
    )

    m: List[AstralBodyPosition] = []
    for interval in range(3):
        pos = moon_position(jd2000 + (interval * 0.5))
        m.append(pos)

    for interval in range(1, 3):
        if m[interval].right_ascension <= m[interval - 1].right_ascension:
            m[interval].right_ascension = m[interval].right_ascension + 2 * pi

    moon_position_window: List[AstralBodyPosition] = [
        replace(m[0]),  # copy m[0]
        AstralBodyPosition(),
        AstralBodyPosition(),
    ]

    rise_time = None
    set_time = None

    # events = []
    for hour in range(24):
        ph = (hour + 1) / 24
        moon_position_window[2].right_ascension = interpolate(
            m[0].right_ascension,
            m[1].right_ascension,
            m[2].right_ascension,
            ph,
        )
        moon_position_window[2].declination = interpolate(
            m[0].declination,
            m[1].declination,
            m[2].declination,
            ph,
        )

        transit_info = moon_transit_event(
            hour, t0, observer.latitude, m[1].distance, moon_position_window
        )
        if isinstance(transit_info, NoTransit):
            moon_position_window[2].distance = transit_info.parallax
        else:
            query_time = datetime.datetime(
                on.year, on.month, on.day, hour, 0, 0, tzinfo=datetime.timezone.utc
            )

            if transit_info.event == "rise":
                event_time = transit_info.when
                event = datetime.datetime(
                    on.year,
                    on.month,
                    on.day,
                    event_time.hour,
                    event_time.minute,
                    0,
                    tzinfo=datetime.timezone.utc,
                )
                if rise_time is None:
                    rise_time = event
                else:
                    rq_diff = (rise_time - query_time).total_seconds()
                    eq_diff = (event - query_time).total_seconds()
                    if set_time is not None:
                        sq_diff = (set_time - query_time).total_seconds()
                    else:
                        sq_diff = 0

                    update_rise_time = sgn(rq_diff) == sgn(eq_diff) and fabs(
                        rq_diff
                    ) > fabs(eq_diff)
                    update_rise_time |= sgn(rq_diff) != sgn(eq_diff) and (
                        set_time is not None and sgn(rq_diff) == sgn(sq_diff)
                    )

                    if update_rise_time:
                        rise_time = event
            elif transit_info.event == "set":
                event_time = transit_info.when
                event = datetime.datetime(
                    on.year,
                    on.month,
                    on.day,
                    event_time.hour,
                    event_time.minute,
                    0,
                    tzinfo=datetime.timezone.utc,
                )
                if set_time is None:
                    set_time = event
                else:
                    sq_diff = (set_time - query_time).total_seconds()
                    eq_diff = (event - query_time).total_seconds()
                    if rise_time is not None:
                        rq_diff = (rise_time - query_time).total_seconds()
                    else:
                        rq_diff = 0

                    update_set_time = sgn(sq_diff) == sgn(eq_diff) and fabs(
                        sq_diff
                    ) > fabs(eq_diff)
                    update_set_time |= sgn(sq_diff) != sgn(eq_diff) and (
                        rise_time is not None and sgn(rq_diff) == sgn(sq_diff)
                    )

                    if update_set_time:
                        set_time = event

        moon_position_window[0].right_ascension = moon_position_window[
            2
        ].right_ascension
        moon_position_window[0].declination = moon_position_window[2].declination
        moon_position_window[0].distance = moon_position_window[2].distance

    return rise_time, set_time


def moonrise(
    observer: Observer,
    date: Optional[datetime.date] = None,
    tzinfo: Union[str, datetime.tzinfo] = datetime.timezone.utc,
) -> Optional[datetime.datetime]:
    if isinstance(tzinfo, str):
        tzinfo = zoneinfo.ZoneInfo(tzinfo)  # type: ignore

    if date is None:
        date = today(tzinfo)  # type: ignore
    elif isinstance(date, datetime.datetime):
        date = date.date()

    info = riseset(date, observer)
    if info[0]:
        rise = info[0].astimezone(tzinfo)  # type: ignore
        rd = rise.date()
        if rd != date:
            if rd > date:
                delta = datetime.timedelta(days=-1)
            else:
                delta = datetime.timedelta(days=1)
            new_date = date + delta
            info = riseset(new_date, observer)
            if info[0]:
                rise = info[0].astimezone(tzinfo)  # type: ignore
                rd = rise.date()
                if rd != date:
                    rise = None
        return rise
    else:
        raise ValueError("Moon never rises on this date, at this location")


def moonset(
    observer: Observer,
    date: Optional[datetime.date] = None,
    tzinfo: Union[str, datetime.tzinfo] = datetime.timezone.utc,
) -> Optional[datetime.datetime]:
    if isinstance(tzinfo, str):
        tzinfo = zoneinfo.ZoneInfo(tzinfo)  # type: ignore

    if date is None:
        date = today(tzinfo)  # type: ignore
    elif isinstance(date, datetime.datetime):
        date = date.date()

    info = riseset(date, observer)
    if info[1]:
        set = info[1].astimezone(tzinfo)  # type: ignore
        sd = set.date()
        if sd != date:
            if sd > date:
                delta = datetime.timedelta(days=-1)
            else:
                delta = datetime.timedelta(days=1)
            new_date = date + delta
            info = riseset(new_date, observer)
            if info[1]:
                set = info[1].astimezone(tzinfo)  # type: ignore
                sd = set.date()
                if sd != date:
                    set = None
        return set
    else:
        raise ValueError("Moon never sets on this date, at this location")


def azimuth(
    observer: Observer,
    at: Optional[datetime.datetime] = None,
) -> Degrees:
    if at is None:
        at = now()

    jd2000 = julianday_2000(at)
    position = moon_position(jd2000)
    lst0: Radians = radians(lmst(at, observer.longitude))
    hourangle: Radians = lst0 - position.right_ascension

    sh = sin(hourangle)
    ch = cos(hourangle)
    sd = sin(position.declination)
    cd = cos(position.declination)
    sl = sin(radians(observer.latitude))
    cl = cos(radians(observer.latitude))

    x = -ch * cd * sl + sd * cl
    y = -sh * cd
    azimuth = degrees(atan2(y, x)) % 360
    return azimuth


def elevation(
    observer: Observer,
    at: Optional[datetime.datetime] = None,
):
    if at is None:
        at = now()

    jd2000 = julianday_2000(at)
    position = moon_position(jd2000)
    lst0: Radians = radians(lmst(at, observer.longitude))
    hourangle: Radians = lst0 - position.right_ascension

    sh = sin(hourangle)
    ch = cos(hourangle)
    sd = sin(position.declination)
    cd = cos(position.declination)
    sl = sin(radians(observer.latitude))
    cl = cos(radians(observer.latitude))

    x = -ch * cd * sl + sd * cl
    y = -sh * cd

    z = ch * cd * cl + sd * sl
    r = sqrt(x * x + y * y)
    elevation = degrees(atan2(z, r))

    return elevation


def zenith(
    observer: Observer,
    at: Optional[datetime.datetime] = None,
):
    return 90 - elevation(observer, at)


def _phase_asfloat(date: datetime.date) -> float:
    jd = julianday(date)
    dt = pow((jd - 2382148), 2) / (41048480 * 86400)
    t = (jd + dt - 2451545.0) / 36525
    t2 = pow(t, 2)
    t3 = pow(t, 3)

    d = 297.85 + (445267.1115 * t) - (0.0016300 * t2) + (t3 / 545868)
    d = radians(d % 360.0)

    m = 357.53 + (35999.0503 * t)
    m = radians(m % 360.0)

    m1 = 134.96 + (477198.8676 * t) + (0.0089970 * t2) + (t3 / 69699)
    m1 = radians(m1 % 360.0)

    elong = degrees(d) + 6.29 * sin(m1)
    elong -= 2.10 * sin(m)
    elong += 1.27 * sin(2 * d - m1)
    elong += 0.66 * sin(2 * d)
    elong = elong % 360.0
    elong = int(elong)
    moon = ((elong + 6.43) / 360) * 28
    return moon


def phase(date: Optional[datetime.date] = None) -> float:
    """Calculates the phase of the moon on the specified date.

    Args:
        date: The date to calculate the phase for. Dates are always in the UTC timezone.
              If not specified then today's date is used.

    Returns:
        A number designating the phase.

        ============  ==============
        0 .. 6.99     New moon
        7 .. 13.99    First quarter
        14 .. 20.99   Full moon
        21 .. 27.99   Last quarter
        ============  ==============
    """

    if date is None:
        date = today()

    moon = _phase_asfloat(date)
    if moon >= 28.0:
        moon -= 28.0
    return moon



import pytest
from astral import LocationInfo
from astral.geocoder import LocationDatabase, database
from astral.location import Location


@pytest.fixture
def test_database() -> LocationDatabase:
    return database()


@pytest.fixture
def london_info() -> LocationInfo:
    # return LocationInfo("London", "England", "Europe/London", 51.50853, -0.12574)
    return LocationInfo("London", "England", "Europe/London", 51.5, -0.1333333)


@pytest.fixture
def london(london_info: LocationInfo) -> Location:
    return Location(london_info)


@pytest.fixture
def new_delhi_info() -> LocationInfo:
    return LocationInfo("New Delhi", "India", "Asia/Kolkata", 28.61, 77.22)


@pytest.fixture
def new_delhi(new_delhi_info: LocationInfo) -> Location:
    return Location(new_delhi_info)


@pytest.fixture
def riyadh_info() -> LocationInfo:
    return LocationInfo("Riyadh", "Saudi Arabia", "Asia/Riyadh", 24.71355, 46.67530)


@pytest.fixture
def riyadh(riyadh_info: LocationInfo) -> Location:
    return Location(riyadh_info)


@pytest.fixture
def wellington_info() -> LocationInfo:
    return LocationInfo(
        "Wellington", "New Zealand", "Pacific/Auckland", -41.33, 174.766666
    )


@pytest.fixture
def wellington(wellington_info: LocationInfo) -> Location:
    return Location(wellington_info)


@pytest.fixture
def tromso_info() -> LocationInfo:
    return LocationInfo("Tromso", "Norway", "CET", 69.6, 18.95)


@pytest.fixture
def tromso(tromso_info: LocationInfo) -> Location:
    return Location(tromso_info)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['astral']

package_data = \
{'': ['*']}

extras_require = \
{':python_version < "3.7"': ['dataclasses'],
 ':python_version < "3.9"': ['backports.zoneinfo'],
 ':sys_platform == "win32"': ['tzdata']}

setup_kwargs = {
    'name': 'astral',
    'version': '3.1',
    'description': 'Calculations for the position of the sun and moon.',
    'long_description': "# Astral\n\nThis is 'astral' a Python module which calculates\n\n- Times for various positions of the sun: dawn, sunrise, solar noon,\nsunset, dusk, solar elevation, solar azimuth and rahukaalam.\n- The phase of the moon.\n\nFor documentation see the <https://astral.readthedocs.io/en/latest/index.html>\n\n## Package Status\n\n![GitHub Workflow Status](https://img.shields.io/github/workflow/status/sffjunkie/astral/astral-test) ![PyPI - Downloads](https://img.shields.io/pypi/dm/astral)\n",
    'author': 'Simon Kennedy',
    'author_email': 'sffjunkie+code@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/sffjunkie/astral',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from pprint import pprint
import datetime

from astral.location import Location
from astral.sun import sun

from almost_equal import datetime_almost_equal


def test_Wellington(wellington: Location):
    dt = datetime.date(2020, 2, 11)
    s = sun(wellington.observer, dt, tzinfo=wellington.tzinfo)
    assert datetime_almost_equal(
        s["sunrise"],
        datetime.datetime(2020, 2, 11, 6, 38, 42, tzinfo=wellington.tzinfo),
    )
    assert datetime_almost_equal(
        s["sunset"],
        datetime.datetime(2020, 2, 11, 20, 31, 00, tzinfo=wellington.tzinfo),
    )



import datetime

from astral import hours_to_time
from astral.sidereal import gmst, lmst


def test_gmst():
    dt = datetime.datetime(1987, 4, 10, 0, 0, 0)
    mean_sidereal_time = gmst(dt)

    t = hours_to_time(mean_sidereal_time / 15)
    assert t.hour == 13
    assert t.minute == 10
    assert t.second == 46
    assert t.microsecond == 366821


def test_gmst_with_time():
    dt = datetime.datetime(1987, 4, 10, 19, 21, 0)
    mean_sidereal_time = gmst(dt)
    t = hours_to_time(mean_sidereal_time / 15)
    assert t.hour == 8
    assert t.minute == 34
    assert t.second == 57
    assert t.microsecond == 89578


def test_local_mean_sidereal_time():
    dt = datetime.datetime(1987, 4, 10, 0, 0, 0)
    mean_sidereal_time = lmst(dt, -0.13)
    assert mean_sidereal_time == 197.693195090862 - 0.13



import datetime

from almost_equal import datetime_almost_equal


class TestDateTimeAlmostEqual:
    """Test the datetime comparison function"""

    def test_equal(self):
        d1 = datetime.datetime(2019, 1, 1)
        d2 = datetime.datetime(2019, 1, 1)

        assert datetime_almost_equal(d1, d2)

    def test_not_equal(self):
        d1 = datetime.datetime(2019, 1, 1)
        d2 = datetime.datetime(2019, 1, 1, 12, 2, 0)

        assert not datetime_almost_equal(d1, d2)

    def test_equal_with_delta(self):
        d1 = datetime.datetime(2019, 1, 1, 12, 0, 0)
        d2 = datetime.datetime(2019, 1, 1, 12, 2, 0)

        assert datetime_almost_equal(d1, d2, 121)
